from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import (
    ChooseCharactorEventArguments, CreateObjectEventArguments, 
    RoundPrepareEventArguments, SkillEndEventArguments, 
    SwitchCharactorEventArguments
)

from ...struct import Cost, ObjectPosition

from ...action import (
    Actions, CreateObjectAction, MakeDamageAction, RemoveObjectAction
)

from ...consts import (
    DamageElementalType, DamageType, IconType, ObjectPositionType, SkillType, 
    WeaponType
)

from ...modifiable_values import (
    DamageElementEnhanceValue, DamageIncreaseValue, DamageMultiplyValue, 
    DamageValue
)
from .base import (
    DefendTeamStatus, ElementalInfusionTeamStatus, ExtraAttackTeamStatus, 
    RoundTeamStatus, TeamStatusBase, UsageTeamStatus
)


class IllusoryBubble_3_3(UsageTeamStatus):
    """
    Team status generated by Mona.
    """
    name: Literal['Illusory Bubble'] = 'Illusory Bubble'
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_MULTIPLY(
        self, value: DamageMultiplyValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageMultiplyValue:
        """
        Double damage when skill damage made.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not this charactor use skill, not modify
            return value
        if self.usage > 0:
            value.damage *= 2
            assert mode == 'REAL'
            self.usage -= 1
        return value


class RainbowBladework_3_6(UsageTeamStatus, ExtraAttackTeamStatus):
    name: Literal['Rainbow Bladework'] = 'Rainbow Bladework'
    version: Literal['3.6'] = '3.6'
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    trigger_skill_type: SkillType | None = SkillType.NORMAL_ATTACK
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    decrease_usage: bool = True


class RainSword_4_2(DefendTeamStatus):
    name: Literal['Rain Sword'] = 'Rain Sword'
    desc: Literal['', 'talent'] = ''
    version: Literal['4.2'] = '4.2'
    usage: int = 2
    max_usage: int = 2
    min_damage_to_trigger: int = 2
    max_in_one_time: int = 1

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.min_damage_to_trigger == 2:
            self.desc = 'talent'

    def renew(self, new_status: 'RainSword_4_2') -> None:
        super().renew(new_status)
        if new_status.min_damage_to_trigger == 2:
            self.desc = 'talent'


class RainbowBladework_3_3(UsageTeamStatus, ExtraAttackTeamStatus):
    name: Literal['Rainbow Bladework'] = 'Rainbow Bladework'
    version: Literal['3.3']
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    trigger_skill_type: SkillType | None = SkillType.NORMAL_ATTACK
    damage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    decrease_usage: bool = True


class RainSword_3_3(RainSword_4_2):
    version: Literal['3.3']
    min_damage_to_trigger: int = 3


class PrayerOfTheCrimsonCrown_3_8(ElementalInfusionTeamStatus, 
                                  RoundTeamStatus,
                                  ExtraAttackTeamStatus):
    name: Literal[
        'Prayer of the Crimson Crown'] = 'Prayer of the Crimson Crown'
    version: Literal['3.8'] = '3.8'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    switch_damage_usage: int = 1
    switch_damage_max_usage: int = 1
    extra_damage_usage: int = 1
    extra_damage_max_usage: int = 1

    # enhance args
    infused_elemental_type: DamageElementalType = DamageElementalType.HYDRO

    # extra attack args
    trigger_skill_type: SkillType | None = SkillType.NORMAL_ATTACK
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    decrease_usage: bool = False

    # switch charactor attack related

    def _attack(
        self, match: Any, usage_type: Literal['SWITCH', 'EXTRA'], 
        damage: int = 1
    ) -> List[MakeDamageAction]:
        """
        attack enemy active charactor
        """
        if usage_type == 'SWITCH':
            if self.switch_damage_usage <= 0:
                # no usage
                return []
            self.switch_damage_usage -= 1
        else:
            assert usage_type == 'EXTRA'
            if self.extra_damage_usage <= 0:
                # no usage
                return []
            self.extra_damage_usage -= 1
        active_charactor = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        return [MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = active_charactor.position,
                    damage = damage,
                    damage_elemental_type = DamageElementalType.HYDRO,
                    cost = Cost()
                )
            ]
        )]

    def event_handler_SWITCH_CHARACTOR(
        self, event: SwitchCharactorEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self switch charactor, perform attack
        """
        if event.action.player_idx != self.position.player_idx:
            # not self switch charactor
            return []
        return self._attack(match, 'SWITCH')

    def event_handler_CHOOSE_CHARACTOR(
        self, event: ChooseCharactorEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self choose charactor (when ally defeated), perform attack
        """
        if event.action.player_idx != self.position.player_idx:
            # not self switch charactor
            return []
        return self._attack(match, 'SWITCH')

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        reset usage
        """
        self.switch_damage_usage = self.switch_damage_max_usage
        self.extra_damage_usage = self.extra_damage_max_usage
        return super().event_handler_ROUND_PREPARE(event, match)

    # renew

    def renew(self, new_status: 'PrayerOfTheCrimsonCrown_3_8') -> None:
        self.switch_damage_max_usage = max(self.switch_damage_max_usage,
                                           new_status.switch_damage_max_usage)
        self.switch_damage_usage = max(self.switch_damage_usage,
                                       new_status.switch_damage_usage)
        self.extra_damage_max_usage = max(self.switch_damage_max_usage,
                                          new_status.extra_damage_max_usage)
        self.extra_damage_usage = max(self.switch_damage_usage,
                                      new_status.extra_damage_usage)
        return super().renew(new_status)

    # enhance and add damage

    def charactor_is_using_right_weapon(
        self, match: Any, position: ObjectPosition
    ) -> bool:
        """
        If position charactor using effected three weapon, return True.
        """
        charactor = match.player_tables[position.player_idx].charactors[
            position.charactor_idx]
        return charactor.weapon_type in [
            WeaponType.SWORD, WeaponType.CLAYMORE, WeaponType.POLEARM]

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self, value: DamageElementEnhanceValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageElementEnhanceValue:
        """
        If weapon type not right, do not enhance.
        """
        if not self.charactor_is_using_right_weapon(match, value.position):
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(
            value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        +1 our normal attack DMG.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not corresponding charactor, do nothing
            return value
        # add damage
        assert mode == 'REAL'
        value.damage += 1
        return value

    # extra attack

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If candace have talent, and our charactor use normal attack, then
        1 hydro damage.
        """
        if event.action.position.player_idx != self.position.player_idx:
            # not our player
            return []
        if event.action.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        charactors = match.player_tables[self.position.player_idx].charactors
        candace_talent = False
        for charactor in charactors:
            if charactor.name == 'Candace' and charactor.talent is not None:
                candace_talent = True
                break
        if not candace_talent:
            # no candace with talent
            return []
        # make damage
        return self._attack(match, 'EXTRA')


class GoldenChalicesBounty_4_2(TeamStatusBase):
    """
    TODO:
    Currently after a dendro core is generated, it removes dendro core and
    create bountiful core. But it will clear existing dendro core. Need to be
    fixed, as dendro core can be triggerd by Abyss Summon.
    """
    name: Literal["Golden Chalice's Bounty"] = "Golden Chalice's Bounty"
    version: Literal['4.2'] = '4.2'
    usage: int = 0
    max_usage: int = 0
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction | CreateObjectAction]:
        """
        If our create dendro core, remove the dendro core and create 
        Bountiful Core.
        """
        if event.action.object_position.player_idx != self.position.player_idx:
            # not our create
            return []
        if (
            event.action.object_position.area != ObjectPositionType.TEAM_STATUS
            or event.action.object_name != 'Dendro Core'
        ):
            # not teams status dendro core
            return []
        # is dendro core, remove it and create bountiful core
        dendro_core = match.player_tables[
            self.position.player_idx].team_status[event.create_idx]
        assert dendro_core.name == 'Dendro Core'
        return [
            RemoveObjectAction(
                object_position = dendro_core.position,
            ),
            CreateObjectAction(
                object_name = 'Bountiful Core',
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.SUMMON,
                    id = 0
                ),
                object_arguments = {}
            )
        ]


register_class(
    IllusoryBubble_3_3 | RainbowBladework_3_6 | RainSword_4_2 | RainSword_3_3
    | PrayerOfTheCrimsonCrown_3_8 | GoldenChalicesBounty_4_2
    | RainbowBladework_3_3
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import RoundEndEventArguments

from ...action import Actions, ChangeObjectUsageAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)
# Summons


class DrunkenMist_3_3(AttackerSummonBase):
    name: Literal['Drunken Mist'] = 'Drunken Mist'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        our_active = match.player_tables[
            self.position.player_idx].get_active_charactor()
        ret.append(MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = our_active.position,
                    damage = -2,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                )
            ],
        ))
        return ret


# Skills


class IcyPaws(ElementalSkillBase):
    name: Literal['Icy Paws'] = 'Icy Paws'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        args = {}
        if self.is_talent_equipped(match):
            # have talent, shield gain one more usage
            args['usage'] = 2
            args['max_usage'] = 2
        return super().get_actions(match) + [
            self.create_team_status('Cat-Claw Shield', args)
        ]


class SignatureMix(ElementalBurstBase):
    name: Literal['Signature Mix'] = 'Signature Mix'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        our_active = match.player_tables[
            self.position.player_idx].get_active_charactor()
        ret.append(MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = our_active.position,
                    damage = -2,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                )
            ],
        ))
        ret.append(self.create_summon('Drunken Mist'))
        return ret


# Talents


class ShakenNotPurred_4_1(SkillTalent):
    name: Literal['Shaken, Not Purred']
    version: Literal['4.1'] = '4.1'
    charactor_name: Literal['Diona'] = 'Diona'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )
    skill: Literal['Icy Paws'] = 'Icy Paws'


class ShakenNotPurred_3_3(ShakenNotPurred_4_1):
    version: Literal['3.3']
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 4
    )


# charactor base


class Diona_3_3(CharactorBase):
    name: Literal['Diona']
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | IcyPaws | SignatureMix
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Kätzlein Style',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            IcyPaws(),
            SignatureMix()
        ]


register_class(
    Diona_3_3 | ShakenNotPurred_3_3 | ShakenNotPurred_4_1 | DrunkenMist_3_3
)



import sys
from src.lpsim.tools import interact_match_with_agents, read_log
from src.lpsim.network import HTTPServer
from src.lpsim.server.match import MatchConfig
import logging


logging.basicConfig(level = logging.INFO)


if __name__ == '__main__':
    log_path = 'logs.json'
    if len(sys.argv) > 1:
        log_path = sys.argv[1]
    log_str = open(log_path).read()
    agents, match = read_log(log_str, use_16_omni = None)
    # for i in range(2):
    #     charactors = match.player_tables[i].player_deck_information.charactors  # noqa: E501
    #     for c in charactors:
    #         c.hp = c.max_hp = 30
    match.config.history_level = 10
    try:
        interact_match_with_agents(agents[0], agents[1], match)
    except Exception:
        print('!!!!! ERROR: play log failed, play to last success log. !!!!!')
    server = HTTPServer(
        decks = ['', ''],
        match_config = MatchConfig(
            check_deck_restriction = False,
            card_number = None,
            max_same_card_number = None,
            charactor_number = None,
            max_round_number = 999,
            random_first_player = False,
            history_level = 10,
        )
    )
    server.match = match

    server.run()



from typing import Any, List, Literal

from ....utils.class_registry import register_class
from ...modifiable_values import DamageValue

from ...summon.base import AttackerSummonBase

from ...event import ReceiveDamageEventArguments, SkillEndEventArguments

from ...action import (
    Actions, ChangeObjectUsageAction, MakeDamageAction
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, 
    ElementalReactionType, FactionType, ObjectPositionType, SkillType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Summons


class CryoCicins_4_1(AttackerSummonBase):
    name: Literal['Cryo Cicins'] = 'Cryo Cicins'
    version: Literal['4.1'] = '4.1'
    usage: int = 2
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1
    renew_type: Literal['ADD'] = 'ADD'
    decrease_only_self_damage: bool = True

    over_maximum: bool = False

    def renew(self, new_status: 'CryoCicins_4_1') -> None:
        if self.usage + new_status.usage > self.max_usage:
            self.over_maximum = True
        return super().renew(new_status)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction | MakeDamageAction]:
        if event.action.position.player_idx != self.position.player_idx:
            # not our player skill made damage, do nothing
            self.over_maximum = False
            return []
        assert event.action.position.area == ObjectPositionType.SKILL
        charactors = match.player_tables[self.position.player_idx].charactors
        charactor = charactors[event.action.position.charactor_idx]
        if charactor.name != 'Fatui Cryo Cicin Mage':
            # not Fatui Cryo Cicin Mage, do nothing
            self.over_maximum = False
            return []
        ret: List[ChangeObjectUsageAction | MakeDamageAction] = []
        if event.action.skill_type == SkillType.NORMAL_ATTACK:
            # change usage
            ret += [ChangeObjectUsageAction(
                object_position = self.position,
                change_usage = 1,
                max_usage = self.max_usage
            )]
            if self.usage + 1 > self.max_usage:
                self.over_maximum = True
        # if over_maximum and charactor has talent, 2 damage to opposite active
        if self.over_maximum and charactor.talent is not None:
            target_charactor = match.player_tables[
                1 - self.position.player_idx].get_active_charactor()
            ret.append(MakeDamageAction(
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.DAMAGE,
                        target_position = target_charactor.position,
                        damage = 2,
                        damage_elemental_type = self.damage_elemental_type,
                        cost = Cost(),
                    )
                ]
            ))
        self.over_maximum = False
        return ret

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        if (
            event.final_damage.target_position.player_idx 
            != self.position.player_idx
            or event.elemental_reaction == ElementalReactionType.NONE
            or event.final_damage.damage_type != DamageType.DAMAGE
        ):
            # not self player receive damage, or no elemental reaction,
            # or not damage
            return []
        target = match.player_tables[
            event.final_damage.target_position.player_idx
        ].charactors[
            event.final_damage.target_position.charactor_idx
        ]
        if (
            self.decrease_only_self_damage
            and target.name != 'Fatui Cryo Cicin Mage'
        ):
            # only decrease in self damage
            return []
        return [ChangeObjectUsageAction(
            object_position = self.position,
            change_usage = -1
        )]


# Skills


def find_cryo_cicins(
    position: ObjectPosition, match: Any, usage: int
) -> Any | None:
    """
    Find cicins based on self position and match. then if with usage will 
    exceed max usage, return cicin position. otherwise, return None.
    """
    summons = match.player_tables[position.player_idx].summons
    for summon in summons:
        if (
            summon.name == 'Cryo Cicins' 
            and summon.usage + usage > summon.max_usage
        ):
            return summon
    return None


class CicinIcicle(ElementalNormalAttackBase):
    name: Literal['Cicin Icicle'] = 'Cicin Icicle'
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = ElementalNormalAttackBase.get_cost(ElementType.CRYO)


class MistySummons(ElementalSkillBase):
    name: Literal['Misty Summons'] = 'Misty Summons'
    version: Literal['4.1'] = '4.1'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon('Cryo Cicins', { 'version': self.version }),
        ]


class BlizzardBranchBlossom(ElementalBurstBase):
    name: Literal['Blizzard, Branch, Blossom'] = 'Blizzard, Branch, Blossom'
    damage: int = 5
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        # max usage, so find regardless of its current usage
        cicin = find_cryo_cicins(self.position, match, 999)
        args = {}
        if cicin is not None:
            args['usage'] = 1 + cicin.usage
            args['max_usage'] = 1 + cicin.usage
        return super().get_actions(match) + [
            self.element_application_self(match, DamageElementalType.CRYO),
            self.create_team_status('Flowing Cicin Shield', args)
        ]


# Talents


class CicinsColdGlare_3_7(SkillTalent):
    name: Literal["Cicin's Cold Glare"]
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Fatui Cryo Cicin Mage'] = 'Fatui Cryo Cicin Mage'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )
    skill: Literal['Misty Summons'] = 'Misty Summons'


# charactor base


class FatuiCryoCicinMage_4_1(CharactorBase):
    name: Literal['Fatui Cryo Cicin Mage']
    version: Literal['4.1'] = '4.1'
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        CicinIcicle | MistySummons | BlizzardBranchBlossom
    ] = []
    faction: List[FactionType] = [
        FactionType.FATUI
    ]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            CicinIcicle(),
            MistySummons(),
            BlizzardBranchBlossom()
        ]


register_class(FatuiCryoCicinMage_4_1 | CryoCicins_4_1 | CicinsColdGlare_3_7)



from typing import Dict, List, Literal

from ....status.charactor_status.base import UsageCharactorStatus

from .....utils.class_registry import register_class
from ....charactor.charactor_base import SkillBase
from ....modifiable_values import CostValue, DamageIncreaseValue
from ....action import (
    Actions, CreateDiceAction, CreateObjectAction, RemoveObjectAction
)
from ....match import Match
from ....event import (
    ReceiveDamageEventArguments, RoundEndEventArguments, 
    RoundPrepareEventArguments, SkillEndEventArguments
)
from ....consts import (
    DamageType, IconType, ObjectPositionType, SkillType, WeaponType
)
from ....struct import Cost
from ....card.equipment.weapon.base import RoundEffectWeaponBase, WeaponBase
from .....utils.desc_registry import DescDictType


class LostPrayerToTheSacredWinds_4_3(WeaponBase):
    name: Literal["Lost Prayer to the Sacred Winds"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number = 3)
    weapon_type: Literal[WeaponType.CATALYST] = WeaponType.CATALYST
    damage_increase: int = 0
    max_damage: int = 2

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[Actions]:
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        self.damage_increase = min(self.damage_increase + 1, self.max_damage)
        return []


class TulaytullahsRemembrance_4_3(RoundEffectWeaponBase):
    name: Literal["Tulaytullah's Remembrance"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number = 3)
    weapon_type: Literal[WeaponType.CATALYST] = WeaponType.CATALYST
    max_usage_per_round: int = 2

    def value_modifier_COST(
        self, value: CostValue, match: Match, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        If self use charged attack, and has usage, decrease one unaligned die 
        cost.
        """
        if self.usage <= 0:
            # not enough usage, not modify
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True, 
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
            source_area = ObjectPositionType.CHARACTOR
        ):
            # not this charactor use skill, or not equipped, not modify
            return value
        skill: SkillBase = match.get_object(value.position)  # type: ignore
        assert skill is not None
        skill_type: SkillType = skill.skill_type
        if skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack, not modify
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        # try decrease any cost
        value.cost.decrease_cost(None)
        if mode == 'REAL':
            self.usage -= 1
        return value


class BeaconOfTheReedSeaStatus_4_3(UsageCharactorStatus):
    # TODO: in official implementation, two status are separated.
    name: Literal["Beacon of the Reed Sea"] = "Beacon of the Reed Sea"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 999

    icon_type: IconType = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        When equipped charactor is using skill, increase the damage.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not current charactor using skill
            return value
        # modify damage
        assert mode == 'REAL'
        value.damage += self.usage
        self.usage = 0
        return value

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        """
        In prepare, remove self
        """
        self.usage = 0
        return self.check_should_remove()


class BeaconOfTheReedSea_4_3(RoundEffectWeaponBase):
    name: Literal["Beacon of the Reed Sea"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number = 3)
    weapon_type: Literal[WeaponType.CLAYMORE] = WeaponType.CLAYMORE
    max_usage_per_round: int = 1  # usage for elemental skill

    receive_damage_usage: int = 1
    max_receive_damage_usage_per_round: int = 1

    def equip(self, match: Match) -> List[Actions]:
        self.usage = self.max_usage_per_round
        self.receive_damage_usage = self.max_receive_damage_usage_per_round
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        self.usage = self.max_usage_per_round
        self.receive_damage_usage = self.max_receive_damage_usage_per_round
        return []

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        if self.receive_damage_usage == 0:
            # already used
            return []
        if not self.position.check_position_valid(
            event.final_damage.target_position, match, player_idx_same = True,
            charactor_idx_same = True, area_same = True,
            source_area = ObjectPositionType.CHARACTOR,
        ):
            # not equipped, or not self receive damage
            return []
        if event.final_damage.damage_type != DamageType.DAMAGE:
            # not damage
            return []
        # create status
        self.receive_damage_usage -= 1
        return [CreateObjectAction(
            object_name = self.name,
            object_position = self.position.set_area(
                ObjectPositionType.CHARACTOR_STATUS
            ),
            object_arguments = {}
        )]

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        if self.usage == 0:
            # already used
            return []
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
            source_area = ObjectPositionType.CHARACTOR
        ):
            # not self, or not equipped
            return []
        skill: SkillBase = match.get_object(
            event.action.position)  # type: ignore
        assert skill is not None
        if skill.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        self.usage -= 1
        return [CreateObjectAction(
            object_name = self.name,
            object_position = self.position.set_area(
                ObjectPositionType.CHARACTOR_STATUS
            ),
            object_arguments = {}
        )]


class PrimordialJadeWingedSpear_4_3(WeaponBase):
    name: Literal["Primordial Jade Winged-Spear"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number = 3)
    weapon_type: Literal[WeaponType.POLEARM] = WeaponType.POLEARM
    max_damage: int = 3

    def equip(self, match: Match) -> List[Actions]:
        self.damage_increase = 1
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        # reset damage increase
        self.damage_increase = 1
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
            source_area = ObjectPositionType.CHARACTOR
        ):
            # not self, or not equipped
            return []
        # increase damage
        self.damage_increase = min(self.damage_increase + 1, self.max_damage)
        return []


class LightOfFoliarIncision_4_3(RoundEffectWeaponBase):
    name: Literal["Light of Foliar Incision"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number = 3)
    weapon_type: Literal[WeaponType.SWORD] = WeaponType.SWORD
    max_usage_per_round: int = 2

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[CreateDiceAction]:
        if self.usage == 0:
            # out of usage
            return []
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
            source_area = ObjectPositionType.CHARACTOR
        ):
            # not self, or not equipped
            return []
        skill: SkillBase = match.get_object(
            event.action.position)  # type: ignore
        assert skill is not None
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        self.usage -= 1
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 1,
            different = True
        )]


desc: Dict[str, DescDictType] = {
    "WEAPON/Lost Prayer to the Sacred Winds": {
        "names": {
            "en-US": "Lost Prayer to the Sacred Winds",
            "zh-CN": "四风原典"
        },
        "descs": {
            "4.3": {
                "en-US": "For each point of \"Bonus DMG\" this card has, character deals +1 DMG.\nEnd Phase: This card gains 1 point of \"Bonus DMG.\" (Max 2 points)\n(Can only be equipped by Catalyst Characters. Each Character can equip at most 1 Weapon.",  # noqa: E501
                "zh-CN": "此牌每有1点「伤害加成」，角色造成的伤害+1。\n结束阶段：此牌累积1点「伤害加成」。（最多累积到2点）\n（「法器」角色才能装备。角色最多装备1件「武器」）"  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Yuandian.png",  # noqa: E501
        "id": 311106
    },
    "WEAPON/Tulaytullah's Remembrance": {
        "names": {
            "en-US": "Tulaytullah's Remembrance",
            "zh-CN": "图莱杜拉的回忆"
        },
        "descs": {
            "4.3": {
                "en-US": "The character deals +1 DMG.\nWhen the character uses a Charged Attack: Spend 1 less Unaligned Element. (Can be triggered up to twice per Round)\n(Only Catalyst Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n角色进行重击时：少花费1个无色元素。（每回合最多触发2次）\n（「法器」角色才能装备。角色最多装备1件「武器」）"  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Tulaidu.png",  # noqa: E501
        "id": 311107
    },
    "CHARACTOR_STATUS/Beacon of the Reed Sea": {
        "names": {
            "en-US": "Beacon of the Reed Sea",
            "zh-CN": "苇海信标"
        },
        "descs": {
            "4.3": {
                "en-US": "Increase DMG dealt by character this Round the next time they deal DMG by the usage of this status.",  # noqa: E501
                "zh-CN": "本回合内，角色下次造成的伤害额外增加该状态的使用次数点。"  # noqa: E501
            }
        },
    },
    "WEAPON/Beacon of the Reed Sea": {
        "names": {
            "en-US": "Beacon of the Reed Sea",
            "zh-CN": "苇海信标"
        },
        "descs": {
            "4.3": {
                "en-US": "The character deals +1 DMG.\nAfter the character uses an Elemental Skill: +1 additional DMG dealt by character this Round the next time they deal DMG. (Once per Round)\nAfter this character takes DMG: +1 additional DMG dealt by character this Round the next time they deal DMG. (Once per Round)\n(Only Claymore Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n角色使用「元素战技」后：本回合内，角色下次造成的伤害额外+1。（每回合1次）\n角色受到伤害后：本回合内，角色下次造成的伤害额外+1。（每回合1次）\n（「双手剑」角色才能装备。角色最多装备1件「武器」）"  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Xinbiao.png",  # noqa: E501
        "id": 311306
    },
    "WEAPON/Primordial Jade Winged-Spear": {
        "names": {
            "en-US": "Primordial Jade Winged-Spear",
            "zh-CN": "和璞鸢"
        },
        "descs": {
            "4.3": {
                "en-US": "The character deals +1 DMG.\nAfter this character uses a Skill: Until the end of this Round, +1 additional DMG provided by this card. (Up to +2 DMG total.)\n(Only Polearm Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n角色使用技能后：直到回合结束前，此牌所提供的伤害加成值额外+1。（最多累积到+2）\n（「长柄武器」角色才能装备。角色最多装备1件「武器」）"  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Hepuyuan.png",  # noqa: E501
        "id": 311407
    },
    "WEAPON/Light of Foliar Incision": {
        "names": {
            "en-US": "Light of Foliar Incision",
            "zh-CN": "裁叶萃光"
        },
        "descs": {
            "4.3": {
                "en-US": "The character deals +1 DMG.\nAfter the character uses a Normal Attack: Create 1 random Elemental Die. (Can be triggered up to twice per Round)\n(Only Sword Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n角色使用「普通攻击」后：生成1个随机类型的元素骰。（每回合最多触发2次）\n（「单手剑」角色才能装备。角色最多装备1件「武器」）"  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Caiye.png",  # noqa: E501
        "id": 311506
    },
}


register_class(
    LostPrayerToTheSacredWinds_4_3 | TulaytullahsRemembrance_4_3
    | BeaconOfTheReedSea_4_3 | PrimordialJadeWingedSpear_4_3
    | LightOfFoliarIncision_4_3 | BeaconOfTheReedSeaStatus_4_3, desc
)



from src.lpsim.agents.interaction_agent import (
    InteractionAgent_V1_0, InteractionAgent
)
from src.lpsim.agents.nothing_agent import NothingAgent
from src.lpsim.server.match import Match, MatchState
from src.lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_usage, get_pidx_cidx, get_test_id_from_command, set_16_omni, 
    check_hp, make_respond, get_random_state
)


def test_rana():
    """
    first: test one round one time, can use imeediately, only elemental skill
    will trigger, can trigger multiple Rana.
    """
    agent_0 = NothingAgent(player_idx = 0)
    agent_1 = InteractionAgent_V1_0(
        version = '1.0',
        player_idx = 1,
        verbose_level = 0,
        commands = [
            'sw_card',
            'choose 0',
            'reroll 1 2 3 4 5 6 7',
            'card 0 0 omni geo',
            'skill 1 dendro dendro omni',
            'card 0 0 hydro hydro',
            'tune pyro 0',
            'card 0 0 dendro dendro',
            'end',
            'reroll 0 1 2 3 4 5 6 7',
            'skill 0 dendro pyro anemo',
            'tune hydro 0',
            'tune hydro 0',
            'skill 2 dendro dendro omni',
            'end',
            'reroll 3 4 5 6 7',
            'skill 1 dendro dendro dendro',
            'card 0 0 dendro dendro',
            'skill 1 dendro dendro dendro',
            'skill 0 dendro geo anemo'
        ],
        only_use_command = True
    )
    match = Match(version = '0.0.1', random_state = get_random_state())
    deck = {
        'name': 'Deck',
        'charactors': [
            {
                'name': 'DendroMobMage',
                'element': 'DENDRO',
                'hp': 99,
                'max_hp': 99,
            },
            {
                'name': 'DendroMobMage',
                'element': 'DENDRO',
            },
            {
                'name': 'ElectroMobMage',
                'element': 'ELECTRO',
            },
        ],
        'cards': [
            {
                'name': 'Rana',
            }
        ] * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            if len(agent_1.commands) == 6:
                # should only have two electro dice
                assert len(match.player_tables[1].dice.colors) == 2
                assert match.player_tables[1].dice.colors == ['ELECTRO', 
                                                              'ELECTRO']
                assert match.player_tables[1].dice.colors[0] == 'ELECTRO'
                assert match.player_tables[1].dice.colors[1] == 'ELECTRO'
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[83, 10, 10], [99, 10, 10]])
    assert len(match.player_tables[1].dice.colors) == 1
    assert match.player_tables[1].dice.colors[0] == 'PYRO'
    assert len(match.player_tables[1].supports) == 4
    for support in match.player_tables[1].supports:
        assert support.name == 'Rana'

    assert match.state != MatchState.ERROR

    """
    second: next one is other people; cannot generate when only one charactor;
    TODO: if overcharged self, will generate next of next?
    """
    agent_0 = InteractionAgent_V1_0(
        version = '1.0',
        player_idx = 0,
        verbose_level = 0,
        commands = [
            'sw_card',
            'choose 2',
            'skill 0 omni omni omni',
            'end',
            'skill 0 omni omni omni',
            'end',
        ],
        only_use_command = True
    )
    agent_1 = InteractionAgent_V1_0(
        version = '1.0',
        player_idx = 1,
        verbose_level = 0,
        commands = [
            'sw_card',
            'choose 0',
            'choose 2',
            'card 0 0 omni omni',
            'card 1 0 omni omni',
            'skill 1 omni omni omni',
            'sw_char 1 omni',
            'end',
            'choose 2',
            'skill 1 omni omni omni',
            'end',
        ],
        only_use_command = True
    )
    match = Match(version = '0.0.1', random_state = get_random_state())
    deck = {
        'name': 'Deck',
        'charactors': [
            {
                'name': 'PyroMobMage',
                'element': 'PYRO',
                'hp': 1,
                'max_hp': 1,
            },
            {
                'name': 'DendroMobMage',
                'element': 'DENDRO',
                'hp': 1,
                'max_hp': 1,
            },
            {
                'name': 'ElectroMobMage',
                'element': 'ELECTRO',
            },
        ],
        'cards': [
            {
                'name': 'Rana',
            }
        ] * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            if len(agent_1.commands) == 4:
                # should be 8 omni and 2 dendro
                assert len(match.player_tables[1].dice.colors) == 10
                omni_num = 8
                dendro_num = 2
                for die in match.player_tables[1].dice.colors:
                    if die == 'OMNI':
                        omni_num -= 1
                    else:
                        assert die == 'DENDRO'
                        dendro_num -= 1
                assert omni_num == 0
                assert dendro_num == 0
            elif len(agent_1.commands) == 1:
                # should be 13 omni
                assert len(match.player_tables[1].dice.colors) == 13
                for die in match.player_tables[1].dice.colors:
                    assert die == 'OMNI'
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert len(agent_0.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[1, 1, 4], [0, 0, 10]])
    assert len(match.player_tables[1].supports) == 2
    for support in match.player_tables[1].supports:
        assert support.name == 'Rana'

    assert match.state != MatchState.ERROR


def test_timmie():
    """
    """
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 0",
            "reroll",
            "card 0 0",
            "end",
            "reroll",
            "card 0 0",
            "card 4 0 6 7",
            "end",
            "TEST 11 record current dice color",
            "reroll",
            "TEST 1 all have 1 omni, compare with previous.0 tm2 1 tm2*3",
            "end",
            "reroll"
        ],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 0",
            "reroll",
            "card 0 0",
            "card 2 0",
            "end",
            "reroll",
            "card 1 0 1",
            "card 1 0",
            "card 2 0",
            "card 2 1",
            "end",
            "reroll",
            "end",
            "TEST 22 record current dice color",
            "reroll",
            "TEST 2 check hand card number",
            "end"
        ],
        only_use_command = True
    )
    match = Match(random_state = get_random_state())
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Fischl
        charactor:Mona
        charactor:Nahida
        Wine-Stained Tricorne*2
        Timmie*2
        Rana*2
        Strategize*2
        # not implement use timmie fill
        Timmie*22
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    last_colors = [[], []]
    while True:
        if match.need_respond(0):
            while True:
                test_id = get_test_id_from_command(agent_0)
                if test_id == 1:
                    current_colors = [
                        match.player_tables[0].dice.colors.copy(),
                        match.player_tables[1].dice.colors.copy()
                    ]
                    assert len(current_colors[0]) == 9
                    assert len(current_colors[1]) == 9
                    assert current_colors[0][0] == 'OMNI'
                    assert current_colors[1][0] == 'OMNI'
                    assert current_colors[0][1:] == last_colors[0]
                    assert current_colors[1][1:] == last_colors[1]
                    tmcount = 0
                    for support in match.player_tables[0].supports:
                        if support.name == 'Timmie':
                            assert support.usage == 2
                            tmcount += 1
                    assert tmcount == 1
                    tmcount = 0
                    for support in match.player_tables[1].supports:
                        if support.name == 'Timmie':  # pragma: no branch
                            assert support.usage == 2
                            tmcount += 1
                    assert tmcount == 3
                elif test_id == 11:
                    last_colors = [
                        match.player_tables[0].dice.colors.copy(),
                        match.player_tables[1].dice.colors.copy()
                    ]
                else:
                    break
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 2:
                    current_colors = [
                        match.player_tables[0].dice.colors.copy(),
                        match.player_tables[1].dice.colors.copy()
                    ]
                    assert len(current_colors[0]) == 9
                    assert len(current_colors[1]) == 11
                    assert current_colors[0][0] == 'OMNI'
                    assert current_colors[1][0] == 'OMNI'
                    assert current_colors[1][1] == 'OMNI'
                    assert current_colors[1][2] == 'OMNI'
                    assert current_colors[0][1:] == last_colors[0]
                    assert current_colors[1][3:] == last_colors[1]
                    for support in match.player_tables[0].supports:
                        assert support.name != 'Timmie'
                    assert len(match.player_tables[1].supports) == 0
                    assert len(match.player_tables[0].hands) == 10
                    assert len(match.player_tables[1].hands) == 10
                elif test_id == 22:
                    last_colors = [
                        match.player_tables[0].dice.colors.copy(),
                        match.player_tables[1].dice.colors.copy()
                    ]
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_liben():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "reroll 1 7",
            "card 0 0",
            "card 0 0",
            "sw_char 1 0",
            "sw_char 0 0",
            "sw_char 1 3",
            "sw_char 0 3",
            "end",
            "TEST 1 p0 usage 2 1",
            "TEST 1 p1 usage 3 3 2 0",
            "reroll",
            "sw_char 1 3",
            "sw_char 0 3",
            "sw_char 1 3",
            "sw_char 0 3",
            "sw_char 1 3",
            "end",
            "TEST 1 p0 usage 3 2",
            "reroll"
        ],
        [
            "sw_card",
            "choose 0",
            "reroll",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "end",
            "reroll",
            "TEST 2 support all 2",
            "TEST 3 hand 7 dice 12 omni 4",
            "end",
            "reroll"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:PyroMobMage
        charactor:ElectroMobMage
        charactor:Noelle
        Liben*30
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            # "TEST 2 support all 2",
            # "TEST 3 hand 7 dice 12 omni 4",
            elif test_id == 1:
                cmd = cmd.strip().split()
                pid = int(cmd[2][1])
                usg = [int(x) for x in cmd[4:]]
                for support, u in zip(match.player_tables[pid].supports, usg):
                    assert support.usage == u
            elif test_id == 2:
                for table in match.player_tables:
                    assert len(table.supports) == 2
            elif test_id == 3:
                table = match.player_tables[1]
                assert len(table.hands) == 7
                assert len(table.dice.colors) == 12
                for color in table.dice.colors[:4]:
                    assert color == 'OMNI'
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_setaria():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "reroll",
            "card 2 0 0",
            "card 3 0 0",
            "tune 0 0",
            "tune 0 0",
            "card 0 0 0",
            "TEST 1 p0 hand 2",
            "card 0 0 0",
            "tune 0 1",
            "TEST 1 p0 hand 1",
            "TEST 2 p0 support usage 2 3 3",
            "tune 0 2",
            "tune 0 3",
            "card 0 0 0",
            "TEST 2 p0 support usage 2 3 3",
            "card 0 0 0",
            "card 0 0 0",
            "card 0 0 0",
            "end",
            "TEST 2 p0 support usage 2 3 3",
            "reroll"
        ],
        [
            "sw_card",
            "choose 0",
            "reroll",
            "sw_char 1 0",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Fischl
        charactor:Mona
        charactor:Nahida
        Setaria*15
        Strategize*15
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                assert len(match.player_tables[pid].hands) == int(cmd[4])
            elif test_id == 2:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                us = [int(x) for x in cmd[5:]]
                assert len(match.player_tables[pid].supports) == len(us)
                for support, u in zip(match.player_tables[pid].supports, us):
                    assert support.usage == u
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_liusu():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 15",
            "card 1 0 14",
            "card 1 0 13",
            "sw_char 1 12",
            "TEST 1 p0 support 1 2 2",
            "sw_char 2 11",
            "sw_char 1 10",
            "TEST 1 p0 support 1 1 2",
            "end",
            "sw_char 0 15",
            "sw_char 1 14"
        ],
        [
            "sw_card",
            "choose 1",
            "TEST 1 p0 support 1 2 2",
            "sw_char 2 15",
            "end",
            "card 1 0 15",
            "sw_char 1 14",
            "TEST 1 p0 support 1 2",
            "TEST 1 p1 support 1",
            "sw_char 0 13",
            "TEST 1 p0 support 1 2",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Electro Hypostasis
        charactor:Klee
        charactor:Keqing
        Liu Su*30
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usage = [int(x) for x in cmd[4:]]
                support = match.player_tables[pidx].supports
                assert len(usage) == len(support)
                for u, s in zip(usage, support):
                    assert u == s.usage
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_tubby():
    cmd_records = [
        [
            "sw_card 4",
            "choose 0",
            "reroll",
            "card 1 0 7 6",
            "card 0 0",
            "card 1 0",
            "reroll 2 3",
            "TEST 1 card 1 cost 2",
            "card 0 0 4 3",
            "card 0 3",
            "TEST 2 p0 dice 4",
            "end",
            "reroll",
            "reroll",
            "card 0 2",
            "card 0 1",
            "end",
            "reroll",
            "card 1 3",
            "end"
        ],
        [
            "sw_card 2 1",
            "choose 0",
            "reroll",
            "end",
            "reroll",
            "card 2 0 6 5",
            "card 1 0 4 3",
            "card 3 0",
            "card 1 0 2 1",
            "end",
            "reroll",
            "card 3 3",
            "reroll",
            "card 3 1 6 5"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Electro Hypostasis
        charactor:Klee
        charactor:Keqing
        Knights of Favonius Library*10
        Tubby*10
        Tenshukaku*10
        Vanarana*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                for req in match.requests:
                    if req.name == 'UseCardRequest' and req.card_idx == 1:
                        assert req.cost.total_dice_cost == 2
            elif test_id == 2:
                assert len(match.player_tables[0].dice.colors) == 4
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_chang_nine():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "card 2 0",
            "skill 0 15 14 13",
            "TEST 1 p0 support 1",
            "TEST 1 p1 support 1",
            "sw_char 4 12",
            "skill 2 11 10 9 8 7",
            "end",
            "card 5 4",
            "card 2 2",
            "card 2 0",
            "end",
            "TEST 1 p0 support 1",
            "TEST 1 p1 support 2",
            "skill 2 15 14 13 12 11"
        ],
        [
            "sw_card 1 2",
            "choose 0",
            "card 2 0",
            "skill 1 15 14 13",
            "sw_char 3 12",
            "TEST 1 p0 support 2",
            "TEST 1 p1 support 2",
            "skill 0 11 10 9",
            "card 1 0",
            "sw_char 1 8",
            "skill 0 7 6 5",
            "TEST 2 p0 card 6",
            "TEST 2 p1 card 5",
            "card 1 1",
            "TEST 1 p1 support 1",
            "sw_char 5 4",
            "skill 0 3 2 1",
            "sw_char 1 0",
            "end",
            "skill 0 15 14 13",
            "TEST 1 p0 support 1",
            "TEST 1 p1 support 2",
            "sw_char 2 12",
            "skill 1 11 10 9",
            "sw_char 3 8",
            "skill 1 7 6 5",
            "end",
            "card 0 0",
            "card 4 0",
            "card 6 0",
            "card 5 0",
            "TEST 1 p1 support 0 0 0 0",
            "card 4 1"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Xingqiu
        charactor:AnemoMobMage
        charactor:Yae Miko
        charactor:Mona
        charactor:Ganyu
        charactor:Klee
        Chang the Ninth*15
        Sweet Madame*15
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usages = [int(x) for x in cmd[4:]]
                supports = match.player_tables[pidx].supports
                assert len(supports) == len(usages)
                for u, s in zip(usages, supports):
                    assert u == s.usage
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                card_num = int(cmd[4])
                assert len(match.player_tables[pidx].hands) == card_num
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_paimon_kujirai():
    cmd_records = [
        [
            "sw_card 1 4",
            "choose 0",
            "reroll",
            "card 3 0 7 6 5",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "end",
            "reroll",
            "TEST 1 p0 support 1 0 0 0",
            "TEST 1 p1 support 1 0",
            "TEST 2 p0 dice 12",
            "TEST 2 p1 dice 12",
            "end",
            "reroll",
            "TEST 1 p0 support 0 0",
            "TEST 1 p1 support 0 0",
            "TEST 2 p0 dice 12",
            "TEST 2 p1 dice 11",
            "card 2 0",
            "card 2 0",
            "sw_char 1 10",
            "end",
            "reroll"
        ],
        [
            "sw_card 2 1",
            "choose 0",
            "reroll",
            "card 0 0",
            "card 0 0",
            "card 0 0 5 4 0",
            "TEST 1 p0 support 2 0 0 0",
            "TEST 1 p1 support 0 0 2",
            "end",
            "reroll",
            "end",
            "reroll",
            "card 1 0",
            "card 0 0",
            "end",
            "reroll",
            "TEST 2 p0 dice 11",
            "TEST 2 p1 dice 12",
            "TEST 1 p0 support 0",
            "TEST 1 p1 support 0 0 0",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Xingqiu
        charactor:AnemoMobMage
        charactor:Yae Miko
        charactor:Mona
        charactor:Ganyu
        charactor:Klee
        Chang the Ninth*15
        Kid Kujirai*15
        Paimon*15
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usages = [int(x) for x in cmd[4:]]
                supports = match.player_tables[pidx].supports
                assert len(supports) == len(usages)
                for u, s in zip(usages, supports):
                    assert u == s.usage
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                dice_num = int(cmd[4])
                assert len(match.player_tables[pidx].dice.colors) == dice_num
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_master_zhang():
    cmd_records = [
        [
            "sw_card 2 4",
            "choose 0",
            "end",
            "end",
            "card 1 0 15",
            "card 1 0 14",
            "TEST 1 card 1 cost 1",
            "card 1 0 13",
            "TEST 2 p0 support 0 0",
            "card 0 0",
            "TEST 1 card 5 cost 1",
            "card 4 0 12",
            "TEST 1 card 4 cost 0",
            "card 0 0 11",
            "card 3 1",
            "TEST 2 p0 support 0 0 0 1",
            "sw_char 1 10",
            "end",
            "card 3 0",
            "TEST 2 p0 support 0 0 1 1",
            "end"
        ],
        [
            "sw_card 1 2",
            "choose 0",
            "end",
            "end",
            "TEST 1 card 2 cost 3",
            "card 2 0 15 14 13",
            "card 4 0 12",
            "card 4 1 11",
            "end",
            "card 5 1",
            "TEST 1 card 6 cost 3",
            "card 5 0 15",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Xingqiu
        charactor:Ganyu
        charactor:Fischl
        Master Zhang*15
        Where Is the Unseen Razor?*15
        King's Squire*15
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usages = [int(x) for x in cmd[4:]]
                supports = match.player_tables[pidx].supports
                assert len(supports) == len(usages)
                for u, s in zip(usages, supports):
                    assert u == s.usage
            elif test_id == 1:
                cmd = cmd.split()
                cid = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == 'UseCardRequest' and req.card_idx == cid:
                        assert req.cost.total_dice_cost == cost
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_katheryne_tian_ellin():
    cmd_records = [
        [
            "sw_card 1 3",
            "choose 0",
            "skill 0 15 14 13",
            "sw_char 2 11",
            "skill 0 10 9 8",
            "card 3 0 7 6",
            "card 3 0 5 4",
            "skill 0 3 2 1",
            "end",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "end",
            "TEST 1 p1 dice 10",
            "TEST 1 p0 dice 16",
            "card 0 0 15",
            "end",
            "sw_char 1 15",
            "sw_char 2 14",
            "skill 2 13 12 11 10",
            "card 0 0 9 8",
            "skill 0 7 6 5",
            "TEST 2 p1 support 0 0 0 1",
            "card 2 0 4 3",
            "TEST 3 skill 0 cost 2",
            "skill 0 2 1",
            "end"
        ],
        [
            "sw_card 1 3",
            "choose 2",
            "skill 0 15 14 13",
            "sw_char 0 11",
            "skill 0 10 9 8",
            "card 2 0 7 6",
            "end",
            "TEST 2 p0 support 1 1",
            "TEST 2 p1 support 1",
            "end",
            "TEST 2 p0 support 1",
            "TEST 4 p0c0 charge 2",
            "TEST 4 p0c1 charge 3",
            "TEST 4 p0c2 charge 2",
            "TEST 4 p1c0 charge 2",
            "TEST 4 p1c1 charge 1",
            "TEST 4 p1c2 charge 1",
            "card 2 0 15",
            "card 2 0 14 13",
            "sw_char 2 12",
            "sw_char 1 11",
            "sw_char 0 10",
            "sw_char 1 9",
            "TEST 2 p0 support 1 1",
            "card 0 0 8 7",
            "card 0 0 6 5",
            "card 3 1 4 3",
            "sw_char 2 2",
            "end",
            "TEST 1 p0 dice 14",
            "TEST 1 p1 dice 16",
            "card 4 0 15 14",
            "skill 1 12 11 10",
            "TEST 3 skill 0 cost 3",
            "TEST 3 skill 1 cost 0",
            "skill 0 9 8 7",
            "TEST 3 skill 0 cost 0",
            "TEST 3 skill 1 cost 0",
            "skill 0",
            "TEST 2 p0 support 1 0 0 1",
            "TEST 2 p1 support 0 0 0 1",
            "TEST 3 skill 0 cost 2",
            "TEST 3 skill 1 cost 2",
            "TEST 3 skill 2 cost 4",
            "end",
            "TEST 3 skill 0 cost 3",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Kaedehara Kazuha
        charactor:Klee
        charactor:Kaeya
        Ellin*10
        Iron Tongue Tian*10
        Katheryne*5
        Katheryne@3.3*5
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].dice.colors) == int(
                    cmd[4])
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            elif test_id == 3:
                sid = int(cmd[3])
                cost = int(cmd[5])
                found = False
                for req in match.requests:
                    if req.name == 'UseSkillRequest' and req.skill_idx == sid:
                        found = True
                        assert req.cost.total_dice_cost == cost
                assert found
            elif test_id == 4:
                pidx, cidx = get_pidx_cidx(cmd)
                assert match.player_tables[pidx].charactors[
                    cidx].charge == int(cmd[4])
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_wagner_timaeus():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 15 14",
            "TEST 1 card 1 cost 1",
            "sw_char 2 13",
            "card 3 0 12 11",
            "end",
            "card 4 0 15 14",
            "TEST 2 p0 support 3 3 2",
            "end",
            "card 2 0 15 14",
            "TEST 1 card 4 cost 3",
            "card 3 0",
            "TEST 1 card 3 cost 0",
            "card 3 1",
            "TEST 2 p0 support 4 4 3 0",
            "end",
            "end",
            "end",
            "card 5 0 15 14",
            "card 0 2 13 12",
            "card 3 3 11 10",
            "card 3 3 9 8",
            "card 3 3 7 6",
            "end"
        ],
        [
            "sw_card 2 1 4",
            "choose 0",
            "card 2 0 15 14",
            "TEST 3 all card cost 0",
            "card 0 1",
            "TEST 2 p1 support 1",
            "TEST 3 all card cost 1",
            "card 0 1 13",
            "end",
            "card 2 0",
            "card 1 1",
            "card 0 1",
            "TEST 2 p1 usage 1",
            "TEST 1 card 0 usage 3",
            "end",
            "TEST 1 card 2 cost 3",
            "end",
            "card 1 0 15 14",
            "end",
            "card 0 1",
            "TEST 2 p1 usage 4 0",
            "end",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Kaedehara Kazuha
        charactor:Klee
        charactor:Kaeya
        Timaeus*10
        Wagner*10
        Gambler's Earrings*10
        Tenacity of the Millelith*5
        Rhythm of the Great Dream*5
        Sacrificial Sword*5
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == 'UseCardRequest' and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            elif test_id == 3:
                cost = int(cmd[-1])
                for req in match.requests:
                    if req.name == 'UseCardRequest':
                        assert req.cost.total_dice_cost == cost
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_chef_mao():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "skill 1 15 14 13",
            "sw_char 0 12",
            "skill 1 11 10 9",
            "card 2 0 8",
            "card 3 0 7",
            "card 1 0",
            "TEST 1 p0 card 4",
            "TEST 3 p0 dice 9",
            "end",
            "sw_char 2 15",
            "sw_char 1 14",
            "sw_char 2 13",
            "card 0 2",
            "TEST 1 p0 card 5",
            "TEST 3 p0 dice 15",
            "card 0 0",
            "card 0 0",
            "TEST 3 p0 dice 15",
            "card 1 0 14",
            "card 1 0 13",
            "end",
            "card 1 3 15",
            "card 0 2",
            "TEST 1 p0 card 3",
            "end",
            "card 3 0 15",
            "card 3 1 14",
            "card 1 1",
            "TEST 1 p0 card 4",
            "end",
            "card 1 1",
            "TEST 1 p0 card 5",
            "card 4 1 15",
            "card 2 0",
            "TEST 1 p0 card 3",
            "end"
        ],
        [
            "sw_card 1 2 3",
            "choose 2",
            "skill 1 15 14 13",
            "card 2 0",
            "TEST 1 p1 card 4",
            "TEST 3 p1 dice 13",
            "skill 1 12 11 10",
            "card 1 0 9",
            "card 1 1",
            "TEST 3 p1 dice 10",
            "TEST 1 p1 card 2",
            "end",
            "sw_char 1 15",
            "sw_char 0 14",
            "skill 0 13 12 11",
            "card 0 0 10",
            "card 0 1",
            "TEST 3 p1 dice 12",
            "TEST 1 p1 card 3",
            "sw_char 1 11",
            "end",
            "end",
            "end",
            "TEST 2 p0 support 0 0 0 0",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Kaedehara Kazuha
        charactor:Klee
        charactor:Kaeya
        Chef Mao*10
        Chef Mao@3.3*10
        Sweet Madame*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[-1])
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            elif test_id == 3:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[
                    pidx].dice.colors) == int(cmd[-1])
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_dunyazard():
    cmd_records = [
        [
            "sw_card 0 1",
            "choose 0",
            "card 0 0 15",
            "TEST 1 card 0 cost 0",
            "TEST 1 card 1 cost 2",
            "TEST 1 card 2 cost 2",
            "end",
            "card 4 0 15 14",
            "end",
            "card 6 0",
            "end",
            "card 6 0",
            "card 4 1 15 14",
            "card 1 3 13 12",
            "end",
            "card 6 3 15 14",
            "end",
            "card 0 3",
            "end",
            "card 0 0",
            "TEST 3 p0 hand 9",
            "TEST 2 p0 support 0 0 1 2",
            "card 1 2",
            "TEST 3 p0 hand 8",
            "card 3 0",
            "TEST 3 p0 hand 8",
            "card 3 0 15 14",
            "end",
            "card 4 3 15 14",
            "card 7 3 13 12",
            "card 3 3",
            "TEST 3 p0 hand 7",
            "card 0 0 11",
            "TEST 3 p0 hand 7",
            "end",
            "end",
            "end",
            "end"
        ],
        [
            "sw_card 1 2",
            "choose 0",
            "TEST 1 card 1 cost 1",
            "card 1 0 15",
            "end",
            "card 5 0",
            "end",
            "card 5 0",
            "end",
            "card 0 0 15 14",
            "card 5 3",
            "end",
            "card 2 0",
            "TEST 2 p1 usage 0 0 1 2",
            "TEST 1 card 0 cost 2",
            "TEST 1 card 1 cost 0",
            "TEST 3 p1 hand 7",
            "card 5 3 15 14",
            "end",
            "card 1 3",
            "end",
            "card 0 0",
            "card 2 3",
            "TEST 3 p1 hand 8",
            "end",
            "card 3 0",
            "card 3 1",
            "end",
            "card 3 0 15 14",
            "end",
            "card 0 3",
            "TEST 3 p1 hand 10",
            "end",
            "TEST 4 p1 deck 11",
            "card 0 3",
            "card 0 3 15 14 13",
            "TEST 4 p1 deck 11",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Kaedehara Kazuha
        charactor:Klee
        charactor:Kaeya
        Dunyarzad*10
        Dunyarzad@3.7*10
        Paimon*10
        NRE@3.3*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == 'UseCardRequest' and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            elif test_id == 3:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[-1])
            elif test_id == 4:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[
                    pidx].table_deck) == int(cmd[-1])
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_xudong():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 2 0 15 14",
            "card 1 1",
            "card 3 1 13 12",
            "card 1 0 11 10",
            "card 3 0",
            "end",
            "choose 1",
            "card 2 0",
            "card 12 0",
            "TEST 1 card 0 cost 1",
            "card 5 0 15 14",
            "card 0 0",
            "TEST 2 p0 usage 0 0 0",
            "end"
        ],
        [
            "sw_card",
            "choose 0",
            "sw_char 2 15",
            "skill 1 14 13 12",
            "skill 1 11 10 9",
            "skill 2 8 7 6 5",
            "skill 1 4 3 2",
            "sw_char 1 1",
            "sw_char 2 0",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Rhodeia of Loch
        charactor:Mona
        charactor:Kaeya
        Xudong*10
        Adeptus' Temptation*10
        Sweet Madame*10
        Lotus Flower Crisp*10
        Teyvat Fried Egg@3.7*10
        # Hanachirusato*10
        # Guardian's Oath*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.config.initial_hand_size = 20
    match.config.max_hand_size = 30
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == 'UseCardRequest' and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_hanachirusato():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 2 15 14 13 12 11",
            "card 3 0",
            "skill 0 10 9 8",
            "skill 0 7 6 5",
            "end",
            "card 2 0",
            "card 2 0",
            "TEST 1 card 0 cost 3",
            "card 2 0 15 14 13 12",
            "TEST 2 p0 usage 3 3 3",
            "card 2 0",
            "TEST 2 p0 usage 3",
            "card 0 0 11",
            "end",
            "TEST 2 p1 usage 3",
            "end"
        ],
        [
            "sw_card",
            "choose 0",
            "card 5 0",
            "card 3 0 15 14 13 12",
            "TEST 1 card 0 cost 3",
            "card 13 0",
            "TEST 2 p1 support 2 0",
            "skill 2 11 10 9 8 7",
            "end",
            "TEST 2 p1 usage 3 1",
            "TEST 2 p0 usage 1",
            "TEST 1 card 0 cost 1",
            "TEST 1 card 1 cost 4",
            "TEST 1 card 4 cost 0",
            "card 4 0",
            "TEST 2 p1 usage 1",
            "skill 2 15 14 13 12 11",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Rhodeia of Loch
        charactor:Mona
        charactor:Noelle
        Hanachirusato*10
        Guardian's Oath*10
        Tenacity of the Millelith*10
        Gambler's Earrings*10
        The Bell*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.config.initial_hand_size = 20
    match.config.max_hand_size = 30
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == 'UseCardRequest' and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_chefmao_dunyarzad():
    cmd_records = [
        [
            "sw_card 2",
            "choose 0",
            "card 2 0 15",
            "TEST 1 p0 hand 4",
            "card 1 0",
            "TEST 1 p0 hand 4",
            "card 3 0",
            "TEST 1 p0 hand 3",
            "skill 0 14 13 12",
            "end",
            "end",
            "card 3 0",
            "TEST 1 p0 hand 7",
            "card 5 0",
            "TEST 1 p0 hand 7",
            "card 3 3",
            "TEST 1 p0 hand 7",
            "card 0 0",
            "TEST 1 p0 hand 8",
            "sw_char 1 15",
            "card 0 0",
            "TEST 1 p0 hand 7",
            "end"
        ],
        [
            "sw_card 2 1",
            "choose 0",
            "card 1 0 15",
            "card 3 0 14",
            "TEST 1 p1 hand 3",
            "card 1 0",
            "card 0 0",
            "TEST 1 p1 hand 3",
            "skill 0 13 15 14",
            "end",
            "end",
            "card 6 0 15",
            "card 0 0",
            "TEST 1 p1 hand 7",
            "card 4 3",
            "TEST 1 p1 hand 6",
            "card 1 0",
            "TEST 1 p1 hand 6",
            "skill 0 14 15 13"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Kaedehara Kazuha
        charactor:Klee
        charactor:Kaeya
        Dunyarzad*10
        Chef Mao*10
        Timmie*10
        Sweet Madame*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx = int(cmd[2][1])
                hand = int(cmd[4])
                assert hand == len(match.player_tables[pidx].hands)
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_yayoi_fix():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "end",
            "end",
            "end",
            "end"
        ],
        [
            "sw_card 2 1",
            "choose 2",
            "card 1 0 15",
            "card 1 1 14 13",
            "end",
            "card 2 2 15",
            "end",
            "card 0 0",
            "end",
            "card 2 1",
            "card 4 2 15 14 13",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.3
        charactor:Yoimiya@3.8
        charactor:Keqing@3.3
        charactor:Dehya@4.1
        Yayoi Nanatsuki*10
        Tenacity of the Millelith*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_timaeus_fix():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 4 0 1",
            "card 5 0 0 1",
            "TEST 1 card 3 cost 0",
            "TEST 1 card 1 cost 0",
            "card 3 0",
            "TEST 1 card 1 cost 1",
            "end"
        ],
        [
            "sw_card 2 1 4",
            "choose 0",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Kaedehara Kazuha
        charactor:Klee
        charactor:Kaeya
        Strategize*10
        Timaeus*10
        Gambler's Earrings*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == 'UseCardRequest' and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_timaeus_fix_2():
    cmd_records = [
        [
            "sw_card 2 1",
            "choose 2",
            "card 3 0",
            "card 2 0 15 14",
            "card 1 0",
            "end",
            "sw_char 1 15",
            "card 0 0 14 13",
            "end"
        ],
        [
            "sw_card 3 2",
            "choose 2",
            "end",
            "card 1 0 15 14",
            "card 1 0",
            "card 1 0",
            "sw_char 1 13"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.2
        charactor:Baizhu
        charactor:Nilou
        charactor:Noelle
        Wagner*10
        Golden House*10
        The Bell*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    # test_rana()
    # test_timmie()
    # test_liben()
    # test_setaria()
    # test_liusu()
    # test_tubby()
    # test_chang_nine()
    # test_paimon_kujirai()
    # test_master_zhang()
    # test_katheryne_tian_ellin()
    # test_wagner_timaeus()
    # test_chef_mao()
    # test_dunyazard()
    # test_xudong()
    # test_hanachirusato()
    # test_chefmao_dunyarzad()
    # test_yayoi_fix()
    test_timaeus_fix_2()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import RoundPrepareEventArguments, SkillEndEventArguments

from ...action import Actions, ChargeAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, SkillType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class ClawAndThunder(ElementalSkillBase):
    name: Literal['Claw and Thunder'] = 'Claw and Thunder'
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = ElementalSkillBase.get_cost(ElementType.ELECTRO)


class LightningFang(ElementalBurstBase):
    name: Literal['Lightning Fang'] = 'Lightning Fang'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_charactor_status('The Wolf Within')
        ]


# Talents
class Awakening_4_2(SkillTalent):
    name: Literal['Awakening']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Razor'] = 'Razor'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
    )
    skill: Literal['Claw and Thunder'] = 'Claw and Thunder'
    usage: int = 1
    max_usage: int = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        reset usage
        """
        self.usage = self.max_usage
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChargeAction]:
        """
        If equipped and use elemental skill, charge one of our electro 
        charactor
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
            source_area = ObjectPositionType.CHARACTOR
        ):
            # not equipped, or this charactor use skill
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        # find charge target
        table = match.player_tables[self.position.player_idx]
        charactors = table.charactors
        active = charactors[table.active_charactor_idx]
        target_idx = -1
        if (
            active.element == ElementType.ELECTRO
            and active.charge < active.max_charge
        ):
            # active electro and not full charge
            target_idx = table.active_charactor_idx
        else:
            # check other charactors
            for chraractor_idx, charactor in enumerate(charactors):
                if (
                    charactor.element == ElementType.ELECTRO
                    and charactor.charge < charactor.max_charge
                    and charactor.is_alive
                ):
                    target_idx = chraractor_idx
                    break
        if target_idx == -1:
            # no target
            return []
        if self.usage <= 0:
            # no usage
            return []
        self.usage -= 1
        # charge target
        return [ChargeAction(
            player_idx = self.position.player_idx,
            charactor_idx = target_idx,
            charge = 1
        )]


# charactor base


class Razor_3_8(CharactorBase):
    name: Literal['Razor']
    version: Literal['3.8'] = '3.8'
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | ClawAndThunder | LightningFang
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Steel Fang',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ClawAndThunder(),
            LightningFang()
        ]


register_class(Razor_3_8 | Awakening_4_2)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DefendSummonBase

from ...modifiable_values import CombatActionValue, DamageIncreaseValue
from ...event import RoundPrepareEventArguments

from ...action import Actions
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE, DamageElementalType, DamageType, DieColor, 
    ElementType, FactionType, ObjectPositionType, PlayerActionLabels, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    PassiveSkillBase, CharactorBase, SkillTalent
)


class Reflection_3_3(DefendSummonBase):
    name: Literal['Reflection'] = 'Reflection'
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    attack_until_run_out_of_usage: bool = False


class MirrorReflectionOfDoom(ElementalSkillBase):
    name: Literal['Mirror Reflection of Doom'] = 'Mirror Reflection of Doom'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_summon('Reflection')]


class StellarisPhantasm(ElementalBurstBase):
    name: Literal['Stellaris Phantasm'] = 'Stellaris Phantasm'
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 3,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        ret.append(self.create_team_status('Illusory Bubble'))
        return ret


class IllusoryTorrent(PassiveSkillBase):
    name: Literal['Illusory Torrent'] = 'Illusory Torrent'
    usage: int = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round prepare, reset usage
        """
        self.usage = 1
        return []

    def value_modifier_COMBAT_ACTION(
        self, value: CombatActionValue, 
        match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> CombatActionValue:
        """
        When combat action is switch, player index is self, switch from mona to
        other people, has usage, and currently is a combat action,
        change the combat action to quick action and decrease usage.
        """
        if value.action_label & PlayerActionLabels.SWITCH.value == 0:
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
            charactor_idx_same = True,
        ):
            return value
        if self.usage <= 0:
            return value
        # as mona is active charactor, current switch is always combat action
        assert value.do_combat_action
        value.do_combat_action = False
        assert mode == 'REAL'
        self.usage -= 1
        return value


class ProphecyOfSubmersion_3_3(SkillTalent):
    name: Literal['Prophecy of Submersion']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Mona'] = 'Mona'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 3,
    )
    skill: Literal['Stellaris Phantasm'] = 'Stellaris Phantasm'

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, 
        match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        If mona is active charactor, and damage triggered hydro reaction,
        which is made by self, increase damage by 2.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, not modify
            return value
        if not self.position.check_position_valid(
            value.position, match,
            source_area = ObjectPositionType.CHARACTOR,  # quipped
            source_is_active_charactor = True,  # active charactor
            player_idx_same = True,  # self damage
        ):
            return value
        if ElementType.HYDRO not in value.reacted_elements:
            # no hydro reaction, not activate
            return value
        value.damage += 2
        return value


class Mona_3_3(CharactorBase):
    name: Literal['Mona']
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        ElementalNormalAttackBase | MirrorReflectionOfDoom | StellarisPhantasm
        | IllusoryTorrent
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT,
    ]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Ripple of Fate',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            MirrorReflectionOfDoom(),
            StellarisPhantasm(),
            IllusoryTorrent(),
        ]


register_class(Mona_3_3 | ProphecyOfSubmersion_3_3 | Reflection_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import (
    Actions, ChangeObjectUsageAction, CreateObjectAction, MakeDamageAction, 
    RemoveObjectAction
)
from ...event import (
    CreateObjectEventArguments, MakeDamageEventArguments, 
    ReceiveDamageEventArguments, RoundPrepareEventArguments, 
    SkillEndEventArguments
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType, DamageType, ElementType, ElementalReactionType, 
    IconType, ObjectPositionType
)

from ...modifiable_values import DamageIncreaseValue, DamageValue
from .base import RoundTeamStatus, ShieldTeamStatus, SwitchActionTeamStatus


class ShrineOfMaya_3_7(RoundTeamStatus):
    name: Literal['Shrine of Maya'] = 'Shrine of Maya'
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        When self is created, check whether to increase the usage of Seed of
        Skandha.
        """
        if event.action.object_name != 'Shrine of Maya':
            # not creating Shrine of Maya, do nothing.
            return []
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        assert active_charactor is not None
        if (
            active_charactor.name == 'Nahida' 
            and active_charactor.talent is not None
        ):
            # talent equipped, check if have electro charactor.
            has_electro_charactor = False
            for charactor in match.player_tables[
                    self.position.player_idx].charactors:
                if charactor.element == ElementType.ELECTRO:
                    has_electro_charactor = True
                    break
            if has_electro_charactor:
                # has electro charactor, add one usage for enemy
                # Seed of Skandha.
                ret = []
                for charactor in match.player_tables[
                        1 - self.position.player_idx].charactors:
                    for status in charactor.status:
                        if status.name == 'Seed of Skandha':
                            position = status.position.set_area(
                                ObjectPositionType.CHARACTOR_STATUS)
                            ret.append(
                                ChangeObjectUsageAction(
                                    object_position = position,
                                    change_usage = 1,
                                )
                            )
                return ret
        return []

    def value_modifier_DAMAGE_INCREASE(
            self, value: DamageIncreaseValue, match: Any,
            mode: Literal['TEST', 'REAL']) -> DamageIncreaseValue:
        """
        +1 on self elemental reaction damage.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, not modify
            return value
        if value.target_position.player_idx == self.position.player_idx:
            # attack self, not activate
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self, not activate
            return value
        if value.element_reaction is not ElementalReactionType.NONE:
            value.damage += 1
        return value


class FloralSidewinder_3_3(RoundTeamStatus):
    """
    Damage made on skill end, but need to check whether dendro reaction made
    before.
    """
    name: Literal['Floral Sidewinder'] = 'Floral Sidewinder'
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    dendro_reaction_made: bool = False
    icon_type: Literal[IconType.ATK_SELF] = IconType.ATK_SELF

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        Check whether dendro reaction made.
        """
        damage = event.final_damage
        if not damage.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not self charactor made damage
            return []
        if ElementType.DENDRO not in damage.reacted_elements:
            # not dendro reaction
            return []
        # mark dendro reaction made
        self.dendro_reaction_made = True
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        if no dendro reaction made, return. otherwise, reset dendro reaction
        made, and if our active charactor use skill, deal 1 dendro damage.
        """
        if not self.dendro_reaction_made:
            # no dendro reaction made
            return []
        # regardless of the following conditions, reset dendro reaction made
        self.dendro_reaction_made = False
        charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        if not event.action.position.check_position_valid(  # pragma: no cover
            charactor.position, match, player_idx_same = True,
            charactor_idx_same = True, source_area = ObjectPositionType.SKILL
        ):
            # not this charactor use skill
            return []  # pragma: no cover
        target = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.DAMAGE,
                        target_position = target.position,
                        damage = 1,
                        damage_elemental_type = DamageElementalType.DENDRO,
                        cost = Cost()
                    )
                ]
            )
        ] + self.check_should_remove()


class AdeptalLegacy_4_1(SwitchActionTeamStatus):
    name: Literal['Adeptal Legacy'] = 'Adeptal Legacy'
    version: Literal['4.1'] = '4.1'
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def _action(self, match: Any) -> List[Actions]:
        """
        attack enemy active charactor and heal our active charactor.
        """
        enemy_active_charactor = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        our_active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        return [
            MakeDamageAction(
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.DAMAGE,
                        target_position = enemy_active_charactor.position,
                        damage = 1,
                        damage_elemental_type = DamageElementalType.DENDRO,
                        cost = Cost()
                    )
                ]
            ),
            MakeDamageAction(
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.HEAL,
                        target_position = our_active_charactor.position,
                        damage = -1,
                        damage_elemental_type = DamageElementalType.HEAL,
                        cost = Cost()
                    )
                ]
            ),
        ]


class PulsingClarity_4_2(RoundTeamStatus):
    name: Literal['Pulsing Clarity'] = 'Pulsing Clarity'
    version: Literal['4.2'] = '4.2'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        ret = super().event_handler_ROUND_PREPARE(event, match)
        return [
            CreateObjectAction(
                object_name = 'Seamless Shield',
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.TEAM_STATUS,
                    id = 0
                ),
                object_arguments = {}
            )
        ] + ret


class SeamlessShield_4_2(ShieldTeamStatus):
    name: Literal['Seamless Shield'] = 'Seamless Shield'
    version: Literal['4.2'] = '4.2'
    usage: int = 1
    max_usage: int = 1
    remove_triggered: bool = False

    def effect_action(self, match: Any) -> MakeDamageAction:
        """
        make damage and heal
        """
        our_active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        opposite_active_charactor = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        ret = MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = opposite_active_charactor.position,
                    damage = 1,
                    damage_elemental_type = DamageElementalType.DENDRO,
                    cost = Cost()
                ),
            ],
        )
        if our_active_charactor.hp > 0:
            # when active charactor has no hp, no need to heal
            ret.damage_value_list.append(
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = our_active_charactor.position,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost()
                ),
            )
        return ret

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        if not self.position.check_position_valid(
            event.action.object_position, match, player_idx_same = True,
            area_same = True,
        ):
            # not create on our area
            return []
        if (
            event.action.object_name != self.name
            or event.create_result != 'RENEW'
        ):
            # not create ourself, or not renew create
            return []
        # renew, make damage and heal
        return [self.effect_action(match)]

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        When damage made, check whether the team status should be removed.
        """
        if self.remove_triggered:
            # already triggered, do nothing
            return []
        ret = self.check_should_remove()
        if len(ret) > 0:
            # should remove, remove self, then make damage and heal
            self.remove_triggered = True
            return [
                self.effect_action(match),
                RemoveObjectAction(object_position = self.position),
            ]
        return list(ret)


register_class(
    ShrineOfMaya_3_7 | FloralSidewinder_3_3 | AdeptalLegacy_4_1 
    | PulsingClarity_4_2 | SeamlessShield_4_2
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import DamageIncreaseValue
from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class GuideToAfterlife(ElementalSkillBase):
    name: Literal['Guide to Afterlife'] = 'Guide to Afterlife'
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        gain charge and create status
        """
        return [
            self.charge_self(1),
            self.create_charactor_status('Paramita Papilio'),
        ]


class SpiritSoother(ElementalBurstBase):
    name: Literal['Spirit Soother'] = 'Spirit Soother'
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Check if hp <= 6. If so, increase damage and heal
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        heal = 2
        if charactor.hp <= 6:
            self.damage = 5
            heal = 3
        damage_action = self.attack_opposite_active(match, self.damage, 
                                                    self.damage_type)
        heal_action = self.attack_self(match, -heal)
        damage_action.damage_value_list += heal_action.damage_value_list
        self.damage = 4
        return [
            self.charge_self(-3),
            damage_action,
        ]

# Talents


class SanguineRouge_3_7(SkillTalent):
    name: Literal['Sanguine Rouge']
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Hu Tao'] = 'Hu Tao'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 2
    )
    skill: Literal['Guide to Afterlife'] = 'Guide to Afterlife'

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return value
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not self use damage skill
            return value
        if value.damage_elemental_type != DamageElementalType.PYRO:
            # not pyro damage
            return value
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if charactor.hp > 6:
            # hp > 6
            return value
        # increase damage
        value.damage += 1
        return value


# charactor base


class HuTao_3_7(CharactorBase):
    name: Literal['Hu Tao']
    version: Literal['3.7'] = '3.7'
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | GuideToAfterlife | SpiritSoother
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Secret Spear of Wangsheng',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            GuideToAfterlife(),
            SpiritSoother()
        ]


register_class(HuTao_3_7 | SanguineRouge_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import CostValue, DamageIncreaseValue, DamageValue

from ...struct import Cost, ObjectPosition

from ...consts import (
    CostLabels, DamageElementalType, DamageType, IconType, ObjectPositionType, 
    SkillType
)

from ...action import (
    ActionTypes, Actions, CreateObjectAction, MakeDamageAction, 
    RemoveObjectAction
)

from ...event import (
    CharactorDefeatedEventArguments, CreateObjectEventArguments, 
    MakeDamageEventArguments, RoundPrepareEventArguments, 
    SkillEndEventArguments
)
from .base import (
    CharactorStatusBase, ElementalInfusionCharactorStatus, 
    PrepareCharactorStatus, RoundCharactorStatus, 
    RoundEndAttackCharactorStatus, ShieldCharactorStatus, UsageCharactorStatus
)


class Riptide_4_1(CharactorStatusBase):
    """
    This status will not deal damages directly, and defeat-regenerate is 
    handled by system handler, which is created by Tartaglia'is passive skill
    when game starts.
    """
    name: Literal['Riptide'] = 'Riptide'
    # As Riptide has changed, all status it related to should have different
    # version. To avoid using wrong version of status, related status have
    # not default value.
    version: Literal['4.1']
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    available_handler_in_trashbin: List[ActionTypes] = [
        ActionTypes.CHARACTOR_DEFEATED
    ]

    def event_handler_CHARACTOR_DEFEATED(
        self, event: CharactorDefeatedEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If defeated charactor has Riptide, the next active charactor will get
        Riptide. As this event triggers later than CHOOSE_CHARACTOR, current
        active charactor is set.
        """
        pidx = event.action.player_idx
        cidx = event.action.charactor_idx
        if (
            pidx != self.position.player_idx 
            or cidx != self.position.charactor_idx
        ):
            # defeated charactor not self charactor
            return []
        return [CreateObjectAction(
            object_name = 'Riptide',
            object_position = ObjectPosition(
                player_idx = pidx,
                area = ObjectPositionType.CHARACTOR_STATUS,
                charactor_idx = match.player_tables[pidx].active_charactor_idx,
                id = 0
            ),
            object_arguments = { 'version': self.version }
        )]


class RangedStance_4_1(CharactorStatusBase):
    name: Literal['Ranged Stance'] = 'Ranged Stance'
    # As Riptide has changed, all status it related to should have different
    # version. To avoid using wrong version of status, related status have
    # not default value.
    version: Literal['4.1']
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS
    usage: int = 1
    max_usage: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When self charactor use charged attack, apply Riptide to target 
        character.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
        ):
            # not self charactor use skill
            return []
        skill = match.get_object(event.action.position)
        if (
            not match.player_tables[self.position.player_idx].charge_satisfied
            or skill.skill_type != SkillType.NORMAL_ATTACK
        ):
            # not charged attack
            return []
        charactor = match.player_tables[
            event.action.target_position.player_idx].charactors[
                event.action.target_position.charactor_idx]
        if charactor.is_defeated:
            # defeated charactor cannot attach Riptide
            return []
        return [CreateObjectAction(
            object_name = 'Riptide',
            object_position = event.action.target_position.set_area(
                ObjectPositionType.CHARACTOR_STATUS),
            object_arguments = { 'version': self.version }
        )]

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        If self charactor gain Melee Stance, remove this status.
        """
        if (
            event.action.object_name == 'Melee Stance'
            and self.position.check_position_valid(
                event.action.object_position, match, player_idx_same = True,
                charactor_idx_same = True, 
                target_area = ObjectPositionType.CHARACTOR_STATUS,
            )
        ):
            # self charactor gain melee stance, remove this status
            return [RemoveObjectAction(
                object_position = self.position
            )]
        return []


class MeleeStance_4_1(ElementalInfusionCharactorStatus,
                      RoundCharactorStatus):
    name: Literal['Melee Stance'] = 'Melee Stance'
    # As Riptide has changed, all status it related to should have different
    # version. To avoid using wrong version of status, related status have
    # not default value.
    version: Literal['4.1']
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS
    usage: int = 2
    max_usage: int = 2
    infused_elemental_type: DamageElementalType = DamageElementalType.HYDRO

    extra_attack_usage: int = 2
    extra_attack_max_usage: int = 2

    def self_skill(self, position: ObjectPosition, match: Any) -> bool:
        """
        Whether position is self skill
        """
        return self.position.check_position_valid(
            position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
        )

    def charge_attack(self, skill_position: ObjectPosition, 
                      match: Any) -> bool:
        """
        Whether position is charged attack
        """
        skill = match.get_object(skill_position)
        return (
            match.player_tables[skill_position.player_idx].charge_satisfied
            and skill.skill_type == SkillType.NORMAL_ATTACK
        )

    def opposite_active_has_riptide(self, match: Any) -> bool:
        """
        Whether opposite active charactor has Riptide
        """
        active = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        for status in active.status:
            if status.name == 'Riptide':
                return True
        return False

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        If target has Riptide, increase damage by 1
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not this charactor use skill, not modify
            return value
        if not self.opposite_active_has_riptide(match):
            # opposite active charactor does not have Riptide
            return value
        # add damage
        value.damage += 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        When use skill on target with Riptide, deal 1 piercing damage to next
        charactor.
        """
        assert len(event.damages) > 0
        # first must be skill damage
        damage = event.damages[0]
        target_position = damage.final_damage.target_position
        if not self.self_skill(damage.final_damage.position, match):
            # not self charactor use skill
            return []
        if self.opposite_active_has_riptide(match):
            # opposite have Riptide, deal 1 piercing damage to next charactor
            next_idx = match.player_tables[
                1 - self.position.player_idx].next_charactor_idx(
                    target_position.charactor_idx)
            if next_idx is not None and self.extra_attack_usage > 0:
                # has next charactor and has usage:
                charactor = match.player_tables[
                    1 - self.position.player_idx].charactors[next_idx]
                self.extra_attack_usage -= 1
                return [MakeDamageAction(
                    damage_value_list = [
                        DamageValue(
                            position = self.position,
                            damage_type = DamageType.DAMAGE,
                            target_position = charactor.position,
                            damage = 1,
                            damage_elemental_type 
                            = DamageElementalType.PIERCING,
                            cost = Cost(),
                        )
                    ]
                )]
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction | MakeDamageAction]:
        """
        When self charactor use charged attack, apply Riptide to target 
        character.
        """
        ret: List[CreateObjectAction | MakeDamageAction] = []

        if not self.self_skill(event.action.position, match):
            # not self charactor use skill
            return []
        # Riptide attach conditions
        if self.charge_attack(event.action.position, match):
            # charged attack
            charactor = match.player_tables[
                event.action.target_position.player_idx].charactors[
                    event.action.target_position.charactor_idx]
            if charactor.is_alive:
                ret += [CreateObjectAction(
                    object_name = 'Riptide',
                    object_position = event.action.target_position.set_area(
                        ObjectPositionType.CHARACTOR_STATUS),
                    object_arguments = { 'version': self.version }
                )]

        return ret

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Reset extra attack usage, and when this status is removed, 
        attach Ranged Stance.
        """
        self.extra_attack_usage = self.extra_attack_max_usage
        ret = super().event_handler_ROUND_PREPARE(event, match)
        if len(ret) > 0:
            # has remove action
            assert ret[0].type == ActionTypes.REMOVE_OBJECT
            ret.append(
                CreateObjectAction(
                    object_name = 'Ranged Stance',
                    object_position = self.position,
                    object_arguments = { 'version': self.version }
                )
            )
        return ret


class Riptide_3_7(Riptide_4_1, RoundCharactorStatus):
    """
    This status will not deal damages directly, and defeat-regenerate is 
    handled by system handler, which is created by Tartaglia'is passive skill
    when game starts.
    """
    name: Literal['Riptide'] = 'Riptide'
    version: Literal['3.7']
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS


class RangedStance_3_7(RangedStance_4_1):
    version: Literal['3.7']


class MeleeStance_3_7(MeleeStance_4_1):
    version: Literal['3.7']


class CeremonialGarment_3_5(RoundCharactorStatus):
    name: Literal['Ceremonial Garment'] = 'Ceremonial Garment'
    version: Literal['3.5'] = '3.5'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        If self use normal attack, increase damage by 1.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this charactor use normal attack, not modify
            return value
        # damage +1
        value.damage += 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self use normal attack, heal all our charactors for 1 HP.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
        ):
            # not self use skill
            return []
        skill = match.get_object(event.action.position)
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        action = MakeDamageAction(
            damage_value_list = []
        )
        charactors = match.player_tables[self.position.player_idx].charactors
        for charactor in charactors:
            if charactor.is_alive:
                action.damage_value_list.append(DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                ))
        return [action]


class HeronShield_3_8(ShieldCharactorStatus, PrepareCharactorStatus):
    name: Literal['Heron Shield'] = 'Heron Shield'
    version: Literal['3.8'] = '3.8'
    charactor_name: Literal['Candace'] = 'Candace'
    skill_name: Literal['Heron Strike'] = 'Heron Strike'

    usage: int = 2
    max_usage: int = 2

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        Do not remove when usage becomes zero
        """
        return []


class Refraction_3_3(RoundCharactorStatus):
    name: Literal['Refraction'] = 'Refraction'
    desc: Literal['', 'talent'] = ''
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2

    is_talent_activated: bool = False
    cost_increase_usage: int = 1
    cost_increase_max_usage: int = 1
    icon_type: Literal[IconType.DEBUFF] = IconType.DEBUFF

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.is_talent_activated:
            self.desc = 'talent'

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.cost_increase_usage = self.cost_increase_max_usage
        return super().event_handler_ROUND_PREPARE(event, match)

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL'],
    ) -> CostValue:
        """
        if talent activated, increase switch cost
        """
        if (
            self.is_talent_activated
            and self.cost_increase_usage > 0
            and value.cost.label & CostLabels.SWITCH_CHARACTOR.value != 0
            and value.position.player_idx == self.position.player_idx
            and value.position.charactor_idx == self.position.charactor_idx
        ):
            # talent activated, switch cost, and this charactor switch to
            # other, increase cost
            value.cost.any_dice_number += 1
            if mode == 'REAL':
                self.cost_increase_usage -= 1
        return value

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        if value.damage_elemental_type != DamageElementalType.HYDRO:
            # not hydro damage
            return value
        if value.is_corresponding_charactor_receive_damage(
            self.position, match
        ):
            # this charactor receive damage
            value.damage += 1
        return value


class TakimeguriKanka_4_1(ElementalInfusionCharactorStatus, 
                          UsageCharactorStatus):
    name: Literal['Takimeguri Kanka'] = 'Takimeguri Kanka'
    version: Literal['4.1'] = '4.1'
    usage: int = 2
    max_usage: int = 2
    infused_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    icon_type: Literal[IconType.ATK_UP_WATER] = IconType.ATK_UP_WATER

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        if value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            assert mode == 'REAL'
            self.usage -= 1
            value.damage += 1
            charactor = match.player_tables[
                self.position.player_idx].charactors[
                    self.position.charactor_idx]
            target = match.player_tables[
                value.target_position.player_idx].charactors[
                    value.target_position.charactor_idx]
            if (
                charactor.talent is not None
                and target.hp <= 6
            ):
                value.damage += 1
        return value


class LingeringAeon_4_2(RoundEndAttackCharactorStatus):
    name: Literal['Lingering Aeon'] = 'Lingering Aeon'
    version: Literal['4.2'] = '4.2'
    usage: int = 1
    max_usage: int = 1
    damage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS


register_class(
    Riptide_4_1 | RangedStance_4_1 | MeleeStance_4_1 
    | Riptide_3_7 | RangedStance_3_7 | MeleeStance_3_7 
    | CeremonialGarment_3_5 | HeronShield_3_8
    | Refraction_3_3 | TakimeguriKanka_4_1 | LingeringAeon_4_2
)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....consts import ObjectPositionType, SkillType

from ....action import ChargeAction

from ....event import SkillEndEventArguments

from ....struct import Cost
from .base import RoundEffectArtifactBase


class ExilesCirclet_3_3(RoundEffectArtifactBase):
    name: Literal["Exile's Circlet"]
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(any_dice_number = 2)
    max_usage_per_round: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChargeAction]:
        """
        If self use elemental burst, charge standby charactors.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, 
            source_area = ObjectPositionType.CHARACTOR,
            target_area = ObjectPositionType.SKILL,
        ):
            # not equipped, or not this charactor use skill
            return []
        if self.usage <= 0:
            # no usage left
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_BURST:
            # not elemental burst
            return []
        # charge
        self.usage -= 1
        ret: List[ChargeAction] = []
        for cid, charactor in enumerate(match.player_tables[
                self.position.player_idx].charactors):
            if (
                cid == self.position.charactor_idx
                or charactor.is_defeated
            ):
                # skip self and defeated charactors
                continue
            ret.append(ChargeAction(
                player_idx = self.position.player_idx,
                charactor_idx = cid,
                charge = 1,
            ))
        return ret


register_class(ExilesCirclet_3_3)
